import java.io.*;
import java.text.DecimalFormat;
import java.util.*;



public class Trading {
	
	/**
	 * Transaction cost
	 * 0% => 0.0
	 * 1.49% => 0.0149
	 * 2 % =>  0.02
	 * 3 % => 0.03
	 * 3.99 % => 0.0399 
	 */
	
	private final double transactionCost = 0.00;
		
	
	
	/**
	 * Use to find the maximum value
	 * 
	 * @param data contains the whole set of data, can be yearly, monthly or weekly etc
	 * @return returns the maximum value as double
	 */
	private double getMax(ArrayList<Double> data){
		if(data.size()==0)
			return 0.0;
		
		double max = data.get(0);
		
		for(int i=1; i<data.size(); i++)
			if(max< data.get(i))
				max = data.get(i);
		
		return max;
	}
	
	/**
	 * Use to find the minimum value
	 * 
	 * @param data contains the whole set of data, can be yearly, monthly or weekly etc
	 * @return returns the minimum value as double
	 */
	private double getMin(ArrayList<Double> data){
		if(data.size()==0)
			return 0.0;
		
		double min = data.get(0);
		
		for(int i=1; i<data.size(); i++)
			if(min>data.get(i))
				min = data.get(i);
		
		return min;
	}
	
	/**
	* Calculates average of array list for a specific range
	* @param data contains the data
	* @param start represents the starting point for the average
	* @param end represents the end point for the average
	* @return returns the average of elements of data from start to end
	*/
	
	private double calcAve(ArrayList<Double> data, int start, int end){
		
		double sum = 0;
		start = Math.max(start, 0);
		end = Math.min(end, data.size()-1);	
		for(int i=start; i<=end; i++)
			sum += data.get(i);
		
		return sum/(end-start+1);
	}
	
	/*
	* Populates the HRLP matrix for Kao and Tate
	*/
	private double [][] calcHRLP(int n){       
             
            double [][] HRLP = new double [4][n+1];
             
            HRLP[1][n] = ((double)(n+1))/2.0; // R(n)
            HRLP[2][n] = 0;   //L(n) = 0
            HRLP[3][n] = 0;  // P(n) = 0 
             
            for(int i=n-1; i>0; i--){
                double t = ((double)(i+1))/((double)(n+1));
                double temp = Math.ceil(t * HRLP[1][i+1]);
                HRLP[0][i] = temp;
                 
                //Calculating R
                temp = HRLP[0][i];
                double t1 = ((double) (n+1))/((double)(2*(i+1))); //ok
                double t2 = ((double)(n+1))/2.0;
                 
                HRLP[1][i] = ((temp-1)/(double)i)*(HRLP[1][i+1] - ( t1* temp ) ) + t2;
                 
                //Calculating L
                temp = HRLP[3][i+1]; // P(i+1)
                t = HRLP[1][i+1];  // R(i+1)
                t1 = ((double)(i+1))/((double)(n+1));
                t2 = Math.floor ( t1 * (t- temp));
                HRLP[2][i] = t2;
                 
                //Calculating P
                temp = HRLP[3][i+1]; // P(i+1)
                t = HRLP[2][i]; // L(i)
                t1 = HRLP[1][i+1]; // R(i+1)
                t2 = ((double)(n+1))/((double)(i+1));
                double t3 = temp+ (t/(double)i)*(t1- temp - ( t2 * (t+1.0)/2.0 )); 
                HRLP[3][i] = t3;
                 
                 
                 
            }
             
            return HRLP;
    }
	
	/**
	 * Runs the Buy RP algorithm for known M and m
	 * RP = SQRT(Mm)
	 * 
	 * @param data contains the whole set of data, can be yearly, monthly or weekly etc
	 * @return returns the OPT and RPMm values
	 */
 	public  double [] buyRPMm(ArrayList<Double> data){
 		
 		double [] result = new double[2];
		
		double m = getMin(data);
		double M = getMax(data);
		
		double RP = Math.sqrt(M*m);
		
		boolean buy = false;
		
		int buyTxs = 0;
		double buyPrice = 0;
		
		
		for(int i=0; i<data.size(); i++){
			
			if(data.get(i)<=RP && buy==false){
				buyPrice = data.get(i);
				buyTxs++;
				buy = true;
				continue;
			}
			else if(data.get(i)<=RP && buy==true){
				buyTxs++;
			}
			else if(i==data.size()-1 && buy==false){
				buyPrice = data.get(i);
				buyTxs++;
			}	
				
		}
		
		result [0] = buyTxs;
		result [1] = buyPrice;
				
		return result;
	}
 	/**
	 * Runs the Sell RP algorithm for known M and m
	 * RP = SQRT(Mm)
	 * 
	 * @param data contains the whole set of data, can be yearly, monthly or weekly etc
	 * @return returns the OPT and RPMm values
	 */
 	public  double [] sellRPMm(ArrayList<Double> data)
	{
 		
 		double [] result = new double[2];
		
		double m = getMin(data);
		double M = getMax(data);
		
		double RP =  Math.sqrt(M*m);
		
		boolean sell = false;
		
		int sellTxs = 0;
		double sellPrice = 0;
		
		
		for(int i=0; i<data.size(); i++){
			
			if(data.get(i)>=RP && sell==false){
				sellPrice = data.get(i);
				sellTxs++;
				sell = true;
				continue;
			}
			else if(data.get(i)>=RP && sell==true){
				sellTxs++;
			}
			else if(i==data.size()-1 && sell==false){
				sellPrice = data.get(i);
				sellTxs++;
			}	
				
		}
		
		result [0] = sellTxs;
		result [1] = sellPrice;
				
		return result;
	}
	
    /**
	 * Runs the Buy KT algorithm 
	 * 
	 * 
	 * @param data contains the whole set of data, can be yearly, monthly or weekly etc
	 * HRLP contains the limits to accept the prices for buy and sell respectively
	 * @return returns the OPT and KT values
	 */
 	public  double [] buyKT(ArrayList<Double> data){
 		
 		double [][] HRLP = calcHRLP(data.size());
 		double [] result = new double[2];
		
		boolean buy = false;
		
		int buyTxs = 0;
		double buyPrice = 0;
		
		for(int i=0; i<data.size(); i++){
			
			int finalRank = 1;
			
			int rank = getRank(data, i);
			
			if(i < data.size() - 2)
				finalRank = (int)HRLP[2][i+2]; //Buy
			
			if(rank<=finalRank && buy==false) { 
				buyPrice = data.get(i);
				buyTxs++;
				buy = true;
				continue;
			}
			else if (rank<=finalRank && buy==true){
				buyTxs++;
			}
			else if(i==data.size()-1 && buy==false){
				buyPrice = data.get(i);
				buyTxs++;
			}
				
		}
				
		result [0] = buyTxs;
		result [1] = buyPrice;
		return result;
	}

    /**
	 * Runs the Sell KT algorithm 
	 * 
	 * 
	 * @param data contains the whole set of data, can be yearly, monthly or weekly etc
	 * HRLP contains the limits to accept the prices for buy and sell respectively
	 * @return returns the OPT and KT values
	 */
 	public  double [] sellKT(ArrayList<Double> data){
 		
 		double [][] HRLP = calcHRLP(data.size());
 		double [] result = new double[2];
		
		boolean sell = false;
		
		int sellTxs = 0;
		double sellPrice = 0;
		
		for(int i=0; i<data.size(); i++){
			
			int finalRank = 1;
			
			int rank = getRank(data, i);
			
			if(i < data.size() - 2)
				finalRank = (int)HRLP[0][i+2]; //Sell
			
			if(rank>=finalRank && sell==false) { 
				sellPrice = data.get(i);
				sellTxs++;
				sell = true;
				continue;
			}
			else if (rank<=finalRank && sell==true){
				sellTxs++;
			}
			else if(i==data.size()-1 && sell==false){
				sellPrice = data.get(i);
				sellTxs++;
			}
				
		}
				
		result [0] = sellTxs;
		result [1] = sellPrice;
		return result;
		
	}
 	
 	
 	/*
 	*  Obtains the rank of a data point
 	*/
 	private int getRank(ArrayList<Double> data, int i){
		int rank = i+1;
		
		for(int j=0; j<i; j++){
			if(data.get(j)>data.get(i))
				rank--;
		}

		return rank;
	}
	
	/*
	* Implements the Variable Length Moving Average algorithm 
	* @param data represents the data set
	* @short represents the length of short term moving average
	* @long represents the length of long term moving average
	* @band represents the band value
	* @return returns the results array
	*/
 	public double [] VLMA(ArrayList<Double> data, int Short, int Long, double band){
		
		double [] results = new double[6];
		
		boolean buy = false;
		double  buyPrice = 0;
		double sellPrice = 0;
		int buyTx = 0;
		int sellTx = 0;
		int completeTxs = 0;
		int positiveTx = 0;
		int negativeTx = 0;
		double returns = 1;
		
		double shortAve = 0;
		double longAve = 0;
		
		
		for(int i=Long-1; i<data.size(); i++){
			shortAve = calcAve(data,i-Short+1,i);
			longAve = calcAve(data, i-Long+1, i);
			
			
			if(shortAve > (1+band)*longAve){
				buyTx++;
				if(!buy){
					buy = true;
					buyPrice = data.get(i);
				
				}
			}
			else if(shortAve < (1-band)*longAve){
					sellTx++;
					if(buy){
						buy = false;
						sellPrice = data.get(i);
						completeTxs++;
						
						double currentReturns =  ((1-transactionCost)*(sellPrice))/((1+transactionCost)*(buyPrice));
						returns  *= currentReturns;
						
						if(currentReturns >= 1)
							positiveTx++;
						else
							negativeTx++;				
					}	
			}
			
		}
		
		results[0] = buyTx;
		results[1] = sellTx;
		results[2] = completeTxs;
		results[3] = positiveTx;
		results[4] = negativeTx;
		results[5] = Math.pow(returns,1.0/completeTxs);

		return results;
	}
	
	/*
	* Implements the Variable Length Moving Average algorithm 
	* @param data represents the data set
	* @periodLength represents the length trading horizon for BH strategy
	* @return returns the results array
	*/
 	public double [] buyAndHold(ArrayList<Double> data, int periodLength){
		
		double [] results = new double[6];
		
		boolean buy = false;
		double  buyPrice = 0;
		double sellPrice = 0;
		int buyTx = 0;
		int sellTx = 0;
		int completeTxs = 0;
		int positiveTx = 0;
		int negativeTx = 0;
		double returns = 1;
		
		
		for(int i=0; i<data.size();){
			
			if(buy==false){
				System.out.print("Buy at : " + i +"\t");
				buy = true;
				buyPrice = data.get(i);
				buyTx++;
				i += periodLength-1;
			}
			else{
				System.out.println("Sell at : " + i +"\t");
				buy = false;
				sellPrice = data.get(i);
				sellTx++;
				completeTxs++;
				double currentReturns =  ((1-transactionCost)*(sellPrice))/((1+transactionCost)*(buyPrice));
				returns  *= currentReturns;
				
				if(currentReturns >= 1)
					positiveTx++;
				else
					negativeTx++;
				
				i++;
			}
		}
		
		System.out.println();
		results[0] = buyTx;
		results[1] = sellTx;
		results[2] = completeTxs;
		results[3] = positiveTx;
		results[4] = negativeTx;
		results[5] = Math.pow(returns,1.0/completeTxs);
		
		
		return results;
	}

	/*
	* Implements the FixedLength Moving Average algorithm 
	* @param data represents the data set
	* @short represents the length of short term moving average
	* @long represents the length of long term moving average
	* @band represents the band value
	* @return returns the results array
	*/
 	public double [] FLMA(ArrayList<Double> data, int Short, int Long, double band){
	
		double [] results = new double[6];
		
		final int holdingDays = 10;
		
		boolean buy = false;
		double  buyPrice = 0;
		double sellPrice = 0;
		int buyTx = 0;
		int sellTx = 0;
		int completeTxs = 0;
		int positiveTx = 0;
		int negativeTx = 0;
		double returns = 1;
		
		double shortAve = 0;
		double longAve = 0;
		
		
		for(int i=Long-1; i<data.size(); ){
			shortAve = calcAve(data,i-Short+1,i);
			longAve = calcAve(data, i-Long+1, i);
			
			
			if(shortAve > (1+band)*longAve){
				buyTx++;
				if(!buy){
					buy = true;
					buyPrice = data.get(i);
					i += holdingDays;
				
				}
				else{
					i++;
				}
			}
			else if(shortAve < (1-band)*longAve){
					sellTx++;
					if(buy){
						buy = false;
						sellPrice = data.get(i);
						completeTxs++;
						
						double currentReturns =  ((1-transactionCost)*(sellPrice))/((1+transactionCost)*(buyPrice));
						returns  *= currentReturns;
						
						if(currentReturns >= 1)
							positiveTx++;
						else
							negativeTx++;	
						
						i += holdingDays;
					}	
					else{
						i++;
					}
			}
			else{
				i++;
			}
			
		}
		
		results[0] = buyTx;
		results[1] = sellTx;
		results[2] = completeTxs;
		results[3] = positiveTx;
		results[4] = negativeTx;
		results[5] = Math.pow(returns,1.0/completeTxs);
	
		return results;
 	}

	/*
	* Implements the RP algorithm of El-Yaniv et al.  
	* @param data represents the data set
	* @periodLength represents the length trading horizon for RP strategy
	* @return returns the results array
	*/
 	public double[] RP(ArrayList<Double> data, int PeriodLength){
 		double [] results = new double[6];
 		
 		int start  = 0;
 		int end    = Math.min(data.size(), (PeriodLength/2) - 1);   //Divide by 2 to get half length for buy period
 		
 		int buyTxs = 0;
 		int sellTxs = 0;
 		int completeTxs = 0;
 		int positiveTxs = 0;
		int negativeTxs = 0;
 		double returns = 1;
 		
 		
 		while(start <= data.size() - PeriodLength){
 			
 			ArrayList<Double> buyData = new ArrayList<Double> (data.subList(start, end));
 			
 			start = end  + 1;
 			end   =  Math.min(data.size(), start + (PeriodLength/2)-1); 
 			
 			ArrayList<Double> sellData = new ArrayList<Double> (data.subList(start, end));
 			
 			double [] buyResult = this.buyRPMm(buyData);
 			double [] sellResult = this.sellRPMm(sellData);
 			
 			buyTxs += buyResult[0];
 			sellTxs += sellResult[0];
 			completeTxs++;
 			
 			double buyPrice = buyResult[1];
 			double sellPrice = sellResult[1];
 			
 			double currentReturns =  ((1-transactionCost)*(sellPrice))/((1+transactionCost)*(buyPrice));
 			returns  *= currentReturns;
 			
 			if(currentReturns >= 1)
 				positiveTxs++;
 			else
 				negativeTxs++;
 				
 				
 			start =  end  + 1;
 			end   =  Math.min(data.size(), start + (PeriodLength/2)-1); 
 			
 		}
 		
 		results[0] = buyTxs;
		results[1] = sellTxs;
		results[2] = completeTxs;
		results[3] = positiveTxs;
		results[4] = negativeTxs;
		results[5] = Math.pow(returns,1.0/completeTxs);
 		
 		
 		return results;
 		
 	}
 	
 	/*
	* Implements the algorithm of Kao and Tate.  
	* @param data represents the data set
	* @periodLength represents the length trading horizon for RP strategy
	* @return returns the results array
	*/
 	public double[] KT(ArrayList<Double> data, int PeriodLength){
 		double [] results = new double[6];
 		
 		int start  = 0;
 		int end    = Math.min(data.size(), (PeriodLength/2) - 1);   //Divide by 2 to get half length for buy period
 		
 		int buyTxs = 0;
 		int sellTxs = 0;
 		int completeTxs = 0;
 		int positiveTxs = 0;
		int negativeTxs = 0;
 		double returns = 1;
 		
 		
 		while(start <= data.size() - PeriodLength){
 			
 			ArrayList<Double> buyData = new ArrayList<Double>(data.subList(start, end));
 			
 			start = end  + 1;
 			end   =  Math.min(data.size(), start + (PeriodLength/2)-1); 
 			
 			ArrayList<Double> sellData = new ArrayList<Double>(data.subList(start, end));
 			
 			double [] buyResult = this.buyKT(buyData);
 			double [] sellResult = this.sellKT(sellData);
 			
 			buyTxs += buyResult[0];
 			sellTxs += sellResult[0];
 			completeTxs++;
 			
 			double buyPrice = buyResult[1];
 			double sellPrice = sellResult[1];
 			
 			double currentReturns =  ((1-transactionCost)*(sellPrice))/((1+transactionCost)*(buyPrice));
 			returns  *= currentReturns;
 			
 			if(currentReturns >= 1)
 				positiveTxs++;
 			else
 				negativeTxs++;
 				
 				
 			start =  end  + 1;
 			end   =  Math.min(data.size(), start + (PeriodLength/2)-1); 
 			
 		}
 		
 		results[0] = buyTxs;
		results[1] = sellTxs;
		results[2] = completeTxs;
		results[3] = positiveTxs;
		results[4] = negativeTxs;
		results[5] = Math.pow(returns,1.0/completeTxs);
 		
 		
 		return results;
 		
 	}
 	/**
	 * Reads data from file and returns the data as ArrayList of type Double
	 *   
	 * @param fileName Name of the file from where data is read
	 * @return returns the data as ArrayList of type Double
	 */
	private ArrayList<Double> readFile(String fileName)
	{
		ArrayList<Double>  data = new ArrayList<Double>();
		try{
		    FileInputStream fstream = new FileInputStream(fileName);
		    // Get the object of DataInputStream
		    DataInputStream in = new DataInputStream(fstream);
		        BufferedReader br = new BufferedReader(new InputStreamReader(in));
		    String strLine;
		    //Read File Line By Line
		    while ((strLine = br.readLine()) != null)   {
		      // Print the content on the console
		    	data.add(Double.parseDouble(strLine));
		    }
		    //Close the input stream
		    in.close();
		    }catch (Exception e){//Catch exception if any
		      System.err.println("Error: " + e.getMessage());
		    }
		
		return data;
		
	}

	/**
	 * Writes the resultant data to the sepcified file
	 *   
	 * @param str  result that is written to the file  
	 * @param file f, the file where the result is written
	 */
	private void writeResult(String str, File f) throws Exception
	{
		 
	      FileOutputStream fop= new FileOutputStream(f);

	      fop.write(str.getBytes());

	      fop.flush();
	      fop.close();
	      System.out.println("The data has been written");
	      
	}
	
	
	public void run()throws Exception{
		
		final int s = 20;
		final int l = 40;
		final double b = 0.02;
		
		final String strVLMA = "VLMA(" + s + "," + l +"," + b + ")"; 
		final String strFLMA = "FLMA(" + s + "," + l +"," + b + ")"; 
		
		
		File f=new File(Output-BU.txt");
		
		//Input file
		// BTC-USD.txt 
		// USD-Euro.txt
		// USD-JPY.txt
		String fileName = "USD-JPY.txt";
		
		

		DecimalFormat df = new DecimalFormat("#.####");
		
		double [] results = new double[6];
		
		String res = "Strartegy \t Buy Txs \t Sell Tx \t Complete Tx \t Positive Returns \t Negative Returns \t Ave-Returns  \n";
		
		ArrayList<Double> data = readFile(fileName);
		

		results = VLMA(data, s, l, b);
		
		res += strVLMA + "\t"+results[0]+"\t"+results[1]+ "\t" + results[2]+"\t"+results[3]+ "\t" + results[4]+"\t"+results[5]+"\n";
		
		
		results = FLMA(data, s, l, b);
		
		res +=  strFLMA + "\t"+results[0]+"\t"+results[1]+ "\t" + results[2]+"\t"+results[3]+ "\t" + results[4]+"\t"+results[5]+"\n";
		
		results = KT(data, l);
		
		res += "KT("+l+") \t"+results[0]+"\t"+results[1]+ "\t" + results[2]+"\t"+results[3]+ "\t" + results[4]+"\t"+results[5]+"\n";
		
		results = RP(data, l);
		
		res += "RP("+l+") \t"+results[0]+"\t"+results[1]+ "\t" + results[2]+"\t"+results[3]+ "\t" + results[4]+"\t"+results[5]+"\n";
		
		
		results = buyAndHold(data, l);
		
		res += "BH("+l+") \t"+results[0]+"\t"+results[1]+ "\t" + results[2]+"\t"+results[3]+ "\t" + results[4]+"\t"+results[5]+"\n";
		
		
		System.out.println(res);
		writeResult(res, f);
	}

}
